#!/usr/bin/env python3
"""
APAM Local Service that creates local user accounts on request by APAMLocal service.
This sits behind AP NGINX and serves files and directories.
"""

from argparse import ArgumentParser
from configparser import ConfigParser
from mimetypes import types_map
from os import path, EX_OSFILE, EX_SOFTWARE

import cherrypy
from create_user import CreateLocalUserAccount


def setup_request_dispatcher(config):
    """
    Setup request dispatchers for types of calls landing on apamlocal
    :param config:
    :return:
    """
    create_user_controller = CreateLocalUserAccount(config)

    # cherrypy internally imports routes package
    dispatcher = cherrypy.dispatch.RoutesDispatcher()

    # Return empty result for ServiceManager Commands
    dispatcher.connect(name='', route='/APAMLocalSecure/CreateLocalUserAccount',
                       controller=create_user_controller, action="create_user_account")

    return dispatcher


def load_apamlocal_ini(filename):
    """
    Load the apamlocal config ini passed in
    :param filename: Name of the file
    :return: ConfigParser object containing apamlocal.ini
    """
    # TODO: Validate Config before loading
    config = ConfigParser()
    config.read(filename)
    return config


def setup_http_listener(config: ConfigParser):
    """
    Setup an HTTP Listener. We can create as many instances of server as we need.
    All will follow cherrypy's engine life-cycle
    :param config: ConfigParser
    :return: cherrpy._cpserver.Server()
    """
    server = cherrypy._cpserver.Server()
    server.socket_host = '0.0.0.0'
    server.socket_port = int(config['HTTPHosting']['Port'])
    server.subscribe()

    return server


def setup_cherrypy(config: ConfigParser):
    """
    Setup CherryPy for starting
    :param config: Config Parser containing apamlocal.ini
    :return: Instance of CherryPY
    """
    dispatcher = setup_request_dispatcher(config)

    cp_config = {
        '/': {
            'request.dispatch': dispatcher,
        }
    }

    # Setup cherrypy
    cherrypy.tree.mount(root=None, config=cp_config)
    cherrypy.server.unsubscribe()

    # No need to setup SSL Listener since NGINX handles it for us.
    # Just configure a HTTP Listener for us.
    setup_http_listener(config)

    cherrypy.engine.signals.subscribe()

    return cherrypy


def start_and_block(cherrypy_engine):
    """
    Start and Block CherryPY server
    :param cherrypy_instance:
    """
    cherrypy_engine.start()
    cherrypy_engine.block()


def main(config_file_name):
    """
    The Main
    :param config_file_name: Location of apamlocal.ini
    """
    apamlocal_config = load_apamlocal_ini(config_file_name)

    # CherryPy uses mimetypes module to determine the content-type.
    types_map['.csv'] = 'text/csv'

    try:
        cherrypy_instance = setup_cherrypy(apamlocal_config)

        # Start Cherry Py main engine and wait for requests from NGINX.
        start_and_block(cherrypy_instance.engine)

    except Exception as e:
        print("ERROR: Running FSSEerver threw exception:", e)
        exit(EX_SOFTWARE)


if __name__ == '__main__':
    parser = ArgumentParser(description="Start AP APAM Local Service")
    parser.add_argument('config_file', metavar='CONFIG.INI', help='Config File ini')

    args = parser.parse_args()
    config_file = args.config_file

    if not path.exists(config_file):
        print("ERROR: apamlocal.ini could not be found")
        from sys import exit

        exit(EX_OSFILE)

    main(config_file)
