#!/usr/bin/env python3
"""
Class for creating a new user
"""

from configparser import ConfigParser
from os import path, makedirs
from pwd import getpwnam
from shutil import move
from subprocess import check_call, CalledProcessError

import cherrypy
from statsd import StatsClient

from apam_retcodes import APAMReturnCode


class CreateLocalUserAccount(object):
    """
    Handle Creating a Local User Account
    """

    def __init__(self, config: ConfigParser):
        """
        Process requests for Data drive D:
        """
        self.config = config
        self.metric = StatsClient('localhost', 8125, prefix='apamlocal')

    @staticmethod
    def index(**kwargs):
        """
        Sometimes an empty GET is done by some watchdogs to see if we are
        accessible.
        """
        return

    @staticmethod
    def user_exists(user_name):
        try:
            getpwnam(user_name)
        except KeyError:
            return False
        else:
            return True

    @staticmethod
    def run_cmd(cmd_param_list):
        print ("run_cmd: ", cmd_param_list)
        try:
            check_call(cmd_param_list)
            return True
        except CalledProcessError as e:
            print("Unable to execute %s.  Failed with %s" % cmd_param_list, e)
            return False


    @staticmethod
    def response_xml(resp):
        return '<AMResult xmlns="http://schemas.datacontract.org/2004/07/Microsoft.Search.Autopilot.Security.AccessManager">' +\
               resp + '</AMResult>'

    def create_user_account(self, *args, **kwargs):
        print ("create_user_account :")

        if not 'userName' in kwargs:
            print ("ERROR: Username not supplied for creating the user.")
            return self.response_xml(APAMReturnCode.InvalidUserName)

        if not 'groups' in kwargs:
            print ("ERROR: Group not supplied for adding the user to.")
            return self.response_xml(APAMReturnCode.InvalidGroupName)

        if not 'expireAfter' in kwargs:
            print("WARNING: Expiry date not supplied, will use 24 hrs.")
            #TODO:
        else:
            expiry_string = kwargs['expireAfter']

        public_key = cherrypy.request.body.read().decode("utf-8")
        if len(public_key) == 0:
            return "ERROR: SSH Public Key is empty."

        user_name = kwargs['userName']
        user_group = kwargs['groups']
        print ("user :", user_name)
        print ("group :", user_group)
        print ("expiry :", expiry_string)
        print("PUBKEY = ", public_key)

        # Is the account already created?. Then change expiry using
        # $ usermod --expiredate YYYY-MM-DD
        #           --key PASS_MAX_DAYS=-1
        #           LOGIN
        # TODO: Add user to admin
        if self.user_exists(user_name):
            if self.run_cmd(['usermod',
                             # TODO: Calculate expiry
                             '-e 2020-10-10',
                             '--lock',
                             '-aG','sudo',
                             user_name]):
                resp = APAMReturnCode.Ok
            else:
                return self.response_xml(APAMReturnCode.UnknownError)
        else:
            # Create an account with expiry 24 hours from now.
            # $ useradd --gid sudo
            #           --expiredate YYYY-MM-DD
            #           --create-home
            #           --key PASS_MAX_DAYS=-1
            #           LOGIN
            # Create an 'at' job
            # $ echo usermod --expiredate 1 --lock LOGIN | at now + 24 hours
            if self.run_cmd(["useradd",
                             "--gid", "sudo",
                             "-e 2020-10-10",
                             "--create-home",
                             "-s", "/bin/bash",
                             user_name]):
                resp = APAMReturnCode.Ok
            else:
                return self.response_xml(APAMReturnCode.UnknownError)

        # Put the public key in authorized_keys
        try:
            pw = getpwnam(user_name)
            home_dir = pw.pw_dir
        except KeyError:
            return self.response_xml(APAMReturnCode.UnknownError)

        ssh_dir = home_dir + "/.ssh"
        if not path.exists(ssh_dir):
            makedirs(ssh_dir)
        auth_keys_new_path = ssh_dir + "/authorized_keys.new"
        auth_keys = open(auth_keys_new_path, 'w')
        auth_keys.write(public_key)
        auth_keys.close()

        auth_keys_path = ssh_dir + "/authorized_keys"
        move(auth_keys_new_path, auth_keys_path);

        # Valid response
        return self.response_xml(resp)
