#!/usr/bin/env python3

import apamlocal
import unittest
from argparse import ArgumentParser
import requests
from tempfile import NamedTemporaryFile, mkdtemp
from shutil import rmtree
from time import strftime, gmtime, time, mktime, localtime
from datetime import datetime, timezone
from configparser import ConfigParser
from random import randint
from pathlib import PosixPath
import calendar


class APAMLocalE2ETest(unittest.TestCase):
    """
    Start APAMLocal on a random port and run a set of tests.
    """

    @classmethod
    def setUpClass(cls):
        test_apamlocal_ini = get_apamlocal_ini()

        # Get a random port
        cls.http_port = str(randint(13100, 13200))

        test_apamlocal_ini['HTTPHosting']['Port'] = cls.http_port

        cls.data_dir = mkdtemp(suffix='_ap_unittest_data')
        test_apamlocal_ini['LINUX']['DataDir'] = cls.data_dir

        cls.app_dir = mkdtemp(suffix='_ap_unittest_app')
        test_apamlocal_ini['LINUX']['AppDir'] = cls.app_dir

        cls.cherrypy = apamlocal.setup_cherrypy(test_apamlocal_ini)

        cls.cherrypy.engine.start()

    @classmethod
    def tearDownClass(cls):
        cls.cherrypy.engine.exit()
        rmtree(cls.data_dir, ignore_errors=True)
        rmtree(cls.app_dir, ignore_errors=True)

    def test_create_user_account(self):
        """
        Test a user account can be created
        """
        url = 'http://localhost:' + self.http_port + r'/APAMLocalSecure/CreateLocalUserAccount'
        query_string = {'userName': 'sriramsh', 'groups': 'Administrators', 'expireAfter': '1.00:0.00:0.00'}
        resp = requests.post(url, data="BLAHBLAHBLAH_PUBLIC_RSA_KEY", timeout=10, params=query_string)
        # resp = <AMResult xmlns="http://schemas.datacontract.org/2004/07/Microsoft.Search.Autopilot.Security.AccessManager">OK</AMResult>


        self.assertEqual(resp.status_code, 200,
                         "Did not return 200")
        expected_resp = b'<AMResult xmlns="http://schemas.datacontract.org/2004/07/Microsoft.Search.Autopilot.Security.AccessManager">OK</AMResult>'
        self.assertEqual(resp.content, expected_resp,
                         "Did not retuen expected string:\n" + str(expected_resp))


"""
Helper Methods
"""


def verify_http_if_modified_header(self, header_str, date_str, expected_ret):
    """
    Test server handles If-Modified-Since http header errors appropriately.
    """
    with NamedTemporaryFile(dir=self.app_dir, prefix='autopilot.ini.') as tmpfile:
        file_name = PosixPath(tmpfile.name).name
        url = 'http://localhost:' + self.http_port + r'/d/app/' + file_name
        headers = {header_str: date_str}

        # if the content didn't change since time, we get 304
        resp = requests.get(url, timeout=10, headers=headers)
        self.assertEqual(resp.status_code, expected_ret,
                         "Expected HTTP Status Code " + str(expected_ret))


def posixtime_to_win_filetime(posix_time_secs):
    # Windows FILETIME = 100 nanoseconds intervals since Jan 1, 1601, 64-bit.
    # Posix time = seconds since Jan 1, 1970

    # Nano Seconds since Jan 1, 1601 to Jan 1, 1970
    epoch_diff_ns = 116444736000000000
    mtime_ns = posix_time_secs * 10000000
    win_file_time_ns = int(mtime_ns) + epoch_diff_ns

    hex_str = hex(win_file_time_ns)[2:].upper()

    # Prefix the hex string with 0 to fill 16 bytes
    file_time_str = ''.join((((16 - len(hex_str)) * '0'), hex_str))

    return file_time_str


def win_filetime_to_posixtime(win_filetime):
    # Windows FILETIME = 100 nanoseconds intervals since Jan 1, 1601, 64-bit.
    # Posix time = seconds since Jan 1, 1970
    ms = int(win_filetime, 16) / 10.

    return datetime.datetime(1601, 1, 1) + datetime.timedelta(microseconds=ms)


def get_apamlocal_ini():
    """
    Return a apamlocal.ini config object
    """
    config = ConfigParser()
    config.read_string("""
            [CherryPy]
            URLPathCaseSensitive=off

            [Certs]
            Directory=./certs
            SSLCert=CCPSCH010261732.crt
            SSLPrivateKey=CCPSCH010261732.key

            [SSLHosting]
            Port=6798

            [HTTPHosting]
            Port=6698

            [LINUX]
            AppDir=/tmp/ap/app
            DataDir=/tmp/ap/data
                      """)
    return config


if __name__ == '__main__':
    parser = ArgumentParser()
    parser.add_argument('-d')

    args = parser.parse_args()
    print(args)
    unittest.main()
