"""
This module provides DMClient class that can be used to send commands to the
Device Manager in the local cluster.

Example Usage:
    dm_client = DMClient(<dm_url>)
    dm_client.send_dm_request("GatewayWatchdogEcho",
                    "Hello World! Test data for the DM Gateway")
"""

import requests
from . import env
from uuid import uuid4
from socket import gethostname
import os


class DMClient():

    def __init__(cls, url, use_certs=True):
        if url != '':
            cls.dm_url = url
        else:
            cls.dm_url = env._get_secure_dm_gateway_url()

        cls.use_certs = use_certs
        cls.hostname = gethostname()


    def _get_dm_headers(self, cmd):
        # DM Custom Header Strings
        dm_header = {
            'Command': "X-APDM-Command",
            'ClientUserName': "X-APDM-ClientUserName",
            'ClientAppName': "X-APDM-ClientAppName",
            'ClientMachineName': "X-APDM-ClientMachineName",
            'ClientEnvironmentName': "X-APDM-ClientEnvironmentName",
            'ClientClusterName': "X-APDM-ClientClusterName",
            'TraceIdentifier': "X-APDM-TraceIdentifier"
        }

        headers = {
            dm_header['Command']: cmd,
            dm_header['TraceIdentifier']: str(uuid4()),
            dm_header['ClientUserName']: 'sriramsh',
            dm_header['ClientAppName']: 'DMClientLinux',
            dm_header['ClientMachineName']: self.hostname,
            dm_header['ClientEnvironmentName']: env.get_environment_name(),
            dm_header['ClientClusterName']: env.get_cluster_name(),
        }

        return headers

    def send_request(self, cmd, payload):
        dm_headers = self._get_dm_headers(cmd)

        machine_function_cert = None
        if self.use_certs is True:
            machine_function_cert = env.get_machine_function_cert()
            if machine_function_cert is None:
                return None

        print('REQUEST sent to : ', self.dm_url)
        self.response = requests.post(self.dm_url,
                                      cert=machine_function_cert,
                                      # don't verify host cert yet
                                      verify=False,
                                      headers=dm_headers,
                                      data=payload)

        print('DM RESPONSE : ', self.response.text)
        return self.response


def main():
    dm_url = "https://dm.autopilot." + env.get_cluster_name() + ".ap.phx.gbl"
    dm_client = DMClient(dm_url)
    dm_client.send_dm_request("GatewayWatchdogEcho",
                              "Hello World! Test data for the DM Gateway")

if __name__ == '__main__':
    main()
