#!/usr/bin/env python3

"""
Provides functions to query AP environment information.
"""

from socket import gethostname
import configparser
import os

#autopilot_ini = ''

__DATA_ROOT__ = '/ap/data'
__APP_ROOT__ = '/ap/app'

def get_data_root():
    return __DATA_ROOT__

def get_app_root():
    return __APP_ROOT__

def set_data_root(data_root):
    global __DATA_ROOT__
    __DATA_ROOT__ = data_root

def set_app_root(app_root):
    global __APP_ROOT__
    __APP_ROOT__ = app_root

def get_linux_bootstrap_ini_path():
    linux_bootstrap_ini_path = get_data_root() + "/LinuxBootstrapData.ini"
    return linux_bootstrap_ini_path

def get_autopilot_ini_path():
    autopilot_ini_path = get_app_root() + "/autopilot.ini"
    return autopilot_ini_path

def get_autopilot_ini_tmp_path():
    autopilot_ini_temp_path = get_app_root() + "/autopilot.ini.tmp"
    return autopilot_ini_temp_path

def get_app_pkg_root():
    return os.path.join(get_app_root(), 'packedimages')

def get_libxdecode_lib_path():
    libxdecode_lib_path = get_app_root() +  "/getandrunservices/libxdecode.so"
    return libxdecode_lib_path

def get_environment_name():
    # Returns AP Environment Name
    autopilot_ini_path = get_autopilot_ini_path()
    config = load_ini_file(autopilot_ini_path)
    return config['FILES']['Environment']

def get_cluster_name():
    # Returns AP Cluster Name
    autopilot_ini_path = get_autopilot_ini_path()
    config = load_ini_file(autopilot_ini_path)
    return config['FILES']['Cluster']

def get_machine_name():
    # Returns AP Machine Name
    return gethostname()

def get_fsserver_http_port():
    # Returns the http port fsserver is listening on.
    autopilot_ini_path = get_autopilot_ini_path()
    config = load_ini_file(autopilot_ini_path)

    if config.has_option('FSSClient', 'FSSPORT'):
        return config['FSSClient']['FSSPORT']
    else:
        return None

def get_fsserver_secure_port():
    # Returns the SSL port fsserver is listening on.
    autopilot_ini_path = get_autopilot_ini_path()
    config = load_ini_file(autopilot_ini_path)

    if config.has_option('FSSClient', 'FSSSecurePORT'):
        return config['FSSClient']['FSSSecurePORT']
    else:
        return None

def get_cert_root():
    return get_data_root() + '/apcommon/certs/'


def _get_secure_dm_gateway_url():
    # Returns Secure DM gateway URL for the cluster
    return 'https://dm.Autopilot.{0}.ap.gbl/DMGateway'.format(get_cluster_name())

def get_machine_cert():
    cert_name = get_machine_name() + '.crt'
    cert_key = get_machine_name() + '.key'

    cert_root = get_data_root() + '/certs/machine/'

    cert_path = cert_root + cert_name
    cert_key_path = cert_root + cert_key

    if not os.path.isfile(cert_path):
        print('ERROR: cert not found: ', cert_path)
        return None

    if not os.path.isfile(cert_key_path):
        print('ERROR: cert key not found: ', cert_key_path)
        return None

    # TODO: Get from common dir. Set SSL Cert Info
    return (cert_path, cert_key_path)

def get_machine_function_cert():
    cert_name = get_machine_name() + '.crt'
    cert_key = get_machine_name() + '.key'

    cert_root = get_data_root() + '/certs/machine_function/'

    cert_path = cert_root + cert_name
    cert_key_path = cert_root + cert_key

    if not os.path.isfile(cert_path):
        print('ERROR: cert not found: ', cert_path)
        return None

    if not os.path.isfile(cert_key_path):
        print('ERROR: cert key not found: ', cert_key_path)
        return None

    # TODO: Get from common dir. Set SSL Cert Info
    return (cert_path, cert_key_path)

#def _load_autopilot_ini(filename):
#    # TODO: Validate Config before loading
#    config = configparser.ConfigParser()
#    config.read(filename)
#    return config

class INISpaceRemover:
    output_file = None
    def __init__( self, new_output_file ):
        self.output_file = new_output_file

    def write( self, what ):
        self.output_file.write( what.replace( " = ", "=" ) )

def load_ini_file(filename):
    # TODO: Validate Config before loading
    config = configparser.ConfigParser()
    config.optionxform = str
    print ("\nReading: " + filename)
    print (config.sections())
    config.read(filename)
    return config

def write_ini_file(filename,config):
    with open(filename, 'w') as configfile:
        print ("\nWriting: " + filename)
        config.write( INISpaceRemover(configfile) )

def get_ini_value(config,section,variable):
    return config[section][variable]

def set_ini_value(config,section,variable,newvalue):
    config[section][variable] = newvalue
    return config

#def init_env(config_file):
#    global autopilot_ini
#
#    if not os.path.exists(config_file):
#        print("ERROR: ini could not be found :", config_file)
#        from sys import exit
#        exit(1)
#
#    autopilot_ini = load_ini_file(config_file)
