import re
import os
from enum import Enum
from pyoverlayfs.filesystem import filesystem
from pyoverlayfs.exceptions import *


class Layer(object): 
    """ Base class for layers."""

    def __init__(self, directory):
        self._directory = directory

    @property
    def directory(self):
        return self._directory

    def validate(self):
        """ Raises exception if layer can't be applied """
        if not filesystem.isdir(self.directory):
            raise InvalidDirectoryException("Not a directory: %s" % self.directory)

    @classmethod
    def _normalize_directory(cls, directory):
        """ normalize/validate directory """
        if directory is "/":
            raise ValueError("Does not support root directory")
        if not os.path.isabs(directory):
            raise ValueError("Must be an absolute path")

        # Strips tailing slashes from path
        return re.sub(r'\/*$', '', directory)

class BaseLayer(Layer):
    """ Represents the terminating layer in assembly. """

    def depth(self):
        """ Base layer has no children, so depth is always 1 """
        return 1

    def mount(self):
        """ Nothing to mount for base layer. Make sure directory exists. """
        self.validate()

    def unmount(self):
        """ Nothing to unmount for base layer."""
        pass

    @property
    def target_directory(self):
        """ Directory where layer is exposed """
        return self._directory

class LayerAssembly(Layer):
    """ Represents all overlay mountpoints and references to lower layers """

    def __init__(self, directory, target_directory, child_layer, mode="ro"):
        self._child_layer = child_layer
        self._target_directory = target_directory
        self._mode = mode
        super().__init__(directory)

    @classmethod
    def create(cls, directories, target_directory):
        """ Prepare a layer assembly consisting of list of directories """
        if len(directories) < 2:
            raise ValueError("Require at least 2 directories in layer")

        target_directory = cls._normalize_directory(target_directory)

        # Build assembly bottom up, starting with first folder
        child_layer = BaseLayer(directory = directories[0])
        for directory in directories[1:-1]:
            directory = cls._normalize_directory(directory)
            child_layer = cls(directory=directory,
                              target_directory=cls._temporary_target_directory(target_directory,
                                                                               child_layer.depth()),
                              child_layer=child_layer)

        # Return the application of the top (final) layer.
        return cls(directory=directories[-1],
                   target_directory=target_directory,
                   child_layer=child_layer,
                   mode="rw") 

    @classmethod
    def load(cls, target_directory):
        return next((layer for layer in cls.list() if layer.target_directory == target_directory))

    @classmethod
    def list(cls):
        layer_assemblies = []
        upper_mountpoints = {}
        lower_mountpoints = {}
        
        output = filesystem.mounts()
        for mount_line in output.splitlines():
            tokens = mount_line.split()
            if(len(tokens) < 2):
                # Skip extraneous newlines
                continue

            print("Mount line: ", mount_line)
            targetdir = tokens[2]
            options = tokens[5].split(',')

            # TODO: could do this in one pass, but this works for now.
            # Extract directories from mount options
            lowerdir = next(filter(None, [re.search('(?<=lowerdir=).*', token) for token in options])).group(0)
            upperdir = next(filter(None, [re.search('(?<=upperdir=).*', token) for token in options])).group(0)
            workdir = next(filter(None, [re.search('(?<=workdir=).*', token) for token in options])).group(0)

            layer = cls(directory=upperdir,
                        child_layer=BaseLayer(directory=lowerdir),
                        target_directory=targetdir)

            # Collapse layers
            if lowerdir in upper_mountpoints.keys():
                # A previous created layer is a child of this layer.
                child_layer = upper_mountpoints[lowerdir]
                layer.child_layer = child_layer
                if child_layer in layer_assemblies:
                    layer_assemblies.remove(child_layer)

            if targetdir in lower_mountpoints.keys():
                # This layer is a child of one or more previously created layers.
                for upper_layer in lower_mountpoints[targetdir]:
                    upper_layer.child_layer = layer
            else:
                # This layer is the top most layer of this assembly so far.
                layer_assemblies.append(layer)
                    
            upper_mountpoints[targetdir] = layer
            if lowerdir not in lower_mountpoints.keys():
                lower_mountpoints[lowerdir] = []
            lower_mountpoints[lowerdir].append(layer)

        return layer_assemblies

    @classmethod
    def _temporary_target_directory(cls, target_directory, layer_level):
        """ Used by intermediate layers to provide a temporary mountpoint name """
        return os.path.join(os.path.dirname(target_directory),
                            "._%s.%d" % (os.path.basename(target_directory), layer_level))

    def mount(self):
        """ Mount child layers then mount this one. """

        try:
            self.validate()
            self.child_layer.mount()
            filesystem.mount_overlay(lower = self.child_layer.target_directory,
                                     upper = self.directory,
                                     target = self.target_directory,
                                     mode = self.mode)
        except Exception as e:
            self.unmount()
            raise e
            
    def unmount(self):
        try:
            filesystem.unmount_overlay(self.target_directory)
        except Exception as e:
            print ("ERROR: Unable to unmount overlay", e)
            # Best effort unmount for now.  TODO: handle specific unmount failures
            pass

        self.child_layer.unmount()

    def layers(self):
        """ Iterator for the stack of child layers """
        current_layer = self
        while(isinstance(current_layer, LayerAssembly)):
            yield current_layer
            current_layer = current_layer.child_layer
        yield current_layer

    def depth(self):
        """ Directory depth of this assembly """
        return 1 + self.child_layer.depth()

    @property
    def mode(self):
        """ Mount option 'ro' or 'rw'. """
        return self._mode

    @property
    def child_layer(self):
        """ A layer applies on top of its child layer """
        return self._child_layer

    @child_layer.setter
    def child_layer(self, value):
        self._child_layer = value

    @property
    def target_directory(self):
        """ Directory where layer is exposed """
        return self._target_directory
