__author__ = "Kevin Tegtmeier"
__copyright__ = "Copyright (C) Microsoft. All rights reserved."

from abc import ABCMeta
from pysystemdwrapper.systemctl_cmd import SystemctlCmd


class Unit:
    """ Represents a systemd unit """
    class UnknownTypeException(Exception):
        pass

    @classmethod
    def __create(self, name, loadstate, state, description):
        """ Factory method for systmed unit types """
        unittype_parts = name.split('.')
        if len(unittype_parts) < 2:
            raise Unit.UnknownTypeException(
                "Expected unit type name of format *.[type], got %s" % name)
        unittype_str = unittype_parts[-1]
        try:
            typeclass = getattr(Unit, unittype_str.title())
        except AttributeError as e:
            raise Unit.UnknownTypeException("Unknown type %s" % name)
        return typeclass(name, loadstate, state, description)

    @classmethod
    def all(self, unittype=None):
        """ Return a list of all systemd units """
        known_units = []

        if unittype is None:
            output = SystemctlCmd.list_units()
        else:
            output = SystemctlCmd.list_units("--type=%s" % unittype)

        for unit_line in output.splitlines():
            unit = self.__parse_unit_line(unit_line)
            if unit is None:
                continue
            known_units.append(unit)
        return known_units

    @classmethod
    def find(self, name):
        return self.__parse_unit_line(SystemctlCmd.list_units(name))

    @classmethod
    def __parse_unit_line(self, unitline):
        sections = str(unitline).split()
        if(len(sections) < 4):
            return None
        return Unit.__create(name=sections[0],
                             loadstate=sections[1],
                             state=sections[3],
                             description=" ".join(sections[4:]))

    class Base:
        """ Abstract base class for the various systemd unit types """
        __metaclass__ = ABCMeta

        def __init__(self, name, loadstate, state, description):
            self.name = name
            self.loadstate = loadstate
            self.state = state
            self.description = description

        def start(self):
            return SystemctlCmd.start_unit(self.name)

        def stop(self):
            return SystemctlCmd.stop_unit(self.name)

        def reset_failed(self):
            return SystemctlCmd.reset_failed_unit(self.name)

        def refresh(self):
            """ Preserve a 'functional core' by making this a value object """
            return Unit.find(self.name)

    class Service(Base):
        pass

    class Socket(Base):
        pass

    class Device(Base):
        pass

    class Mount(Base):
        pass

    class Automount(Base):
        pass

    class Swap(Base):
        pass

    class Target(Base):
        pass

    class Path(Base):
        pass

    class Timer(Base):
        pass

    class Snapshot(Base):
        pass

    class Slice(Base):
        pass

    class Scope(Base):
        pass
