#!/usr/bin/env python3

import fsserver
import unittest
from argparse import ArgumentParser
from requests import get
from tempfile import NamedTemporaryFile, mkdtemp
from shutil import rmtree
from time import strftime, gmtime, time, mktime, localtime
from datetime import datetime, timezone
from configparser import ConfigParser
from random import randint
from pathlib import PosixPath
import calendar


class FSServerE2ETest(unittest.TestCase):
    """
    Start FSServer on a random port and run a set of tests.
    """

    @classmethod
    def setUpClass(cls):
        test_fsserver_ini = get_fsserver_ini()

        # Get a random port
        cls.http_port = str(randint(13100, 13200))

        test_fsserver_ini['HTTPHosting']['Port'] = cls.http_port

        cls.data_dir = mkdtemp(suffix='_ap_unittest_data')
        test_fsserver_ini['LINUX']['DataDir'] = cls.data_dir

        cls.app_dir = mkdtemp(suffix='_ap_unittest_app')
        test_fsserver_ini['LINUX']['AppDir'] = cls.app_dir

        cls.cherrypy = fsserver.setup_cherrypy(test_fsserver_ini)

        cls.cherrypy.engine.start()

    @classmethod
    def tearDownClass(cls):
        cls.cherrypy.engine.exit()
        rmtree(cls.data_dir, ignore_errors=True)
        rmtree(cls.app_dir, ignore_errors=True)

    def test_data_file_read(self):
        # TEST_THIS_COMMENT_EXISTS
        pass

    def test_dir_list_child_dir(self):
        pass

    def test_dir_listing_win_path(self):
        for drive_alias in ['d', 'drived', 'DriveD', 'driveD', 'DRIVED']:
            for path_root in ['\\app', '\\data']:
                url = 'http://localhost:' + self.http_port + r'/' + drive_alias + path_root
                with self.subTest(url=url):

                    # Create a tmp file and check for the same in dir listing
                    if path_root == '\\app':
                        with NamedTemporaryFile(dir=self.app_dir) as tmpfile:
                            resp = get(url, timeout=10)
                            file_name = PosixPath(tmpfile.name).name
                            verify_dir_listing(self, resp, file_name)
                    else:
                        with NamedTemporaryFile(dir=self.data_dir) as tmpfile:
                            resp = get(url, timeout=10)
                            file_name = PosixPath(tmpfile.name).name
                            verify_dir_listing(self, resp, file_name)

    def test_dir_listing_posix_path(self):
        for drive_alias in ['d', 'drived', 'DriveD', 'driveD', 'DRIVED']:
            for path_root in ['/app', '/data']:
                url = 'http://localhost:' + self.http_port + r'/' + drive_alias + path_root
                with self.subTest(url=url):

                    # Create a tmp file and check for the same in dir listing
                    if path_root == '/app':
                        with NamedTemporaryFile(dir=self.app_dir) as tmpfile:
                            resp = get(url, timeout=10)
                            file_name = PosixPath(tmpfile.name).name
                            verify_dir_listing(self, resp, file_name)
                    else:
                        with NamedTemporaryFile(dir=self.data_dir) as tmpfile:
                            resp = get(url, timeout=10)
                            file_name = PosixPath(tmpfile.name).name
                            verify_dir_listing(self, resp, file_name)

    def test_non_existent_dir_listing(self):
        """
        Test a non-existent dir listing returns 404
        """
        url = 'http://localhost:' + self.http_port + r'/d\data\NON_EXISTENT'
        resp = get(url, timeout=10)
        self.assertEqual(resp.status_code, 404,
                         "Expected HTTP Status Code 404")

    def test_handling_if_modified_since_304(self):
        """
        Test server handles If-Modified-Since header appropriately.
        If the file has not been modified since the time specified, 304 will be
        returned.
        """
        print("If-Modified-Since: Testing if 200 returned on later date.")
        later_time = gmtime(time() + 600)
        later_time_str = strftime('%a, %d %b %Y %H:%M:%S GMT', later_time)
        verify_http_if_modified_header(self,
                                       'If-Modified-Since',
                                       later_time_str, 304)

        print("If-Modified-Since: Testing if MAX date is accepted.")
        max_time = datetime(9999, 12, 31, 23, 59, 59, 999999)
        max_time_str = max_time.strftime('%a, %d %b %Y %H:%M:%S GMT')
        verify_http_if_modified_header(self,
                                       'If-Modified-Since',
                                       max_time_str, 304)

    def test_handling_if_unmodified_since_412(self):
        """
        Test server handles If-Unmodified-Since header appropriately.
        If the file has been modified since the time specified, 412 will be
        returned.
        """
        print("If-Unmodified-Since: Testing if 200 returned on later date.")
        prev_time = gmtime(time() - 10)
        prev_time_str = strftime('%a, %d %b %Y %H:%M:%S GMT', prev_time)
        verify_http_if_modified_header(self,
                                       'If-Unmodified-Since',
                                       prev_time_str, 412)

        print("If-Unmodified-Since: Testing if Unix epoch MIN date is accepted.")
        min_time_win = datetime(1601, 1, 1)
        min_time_str = min_time_win.strftime('%a, %d %b %Y %H:%M:%S GMT')
        verify_http_if_modified_header(self,
                                       'If-Unmodified-Since',
                                       min_time_str, 412)

        print("If-Unmodified-Since: Testing if Win min start date is accepted.")
        min_time_unix = datetime(1970, 1, 1)
        min_time_str = min_time_unix.strftime('%a, %d %b %Y %H:%M:%S GMT')
        verify_http_if_modified_header(self,
                                       'If-Unmodified-Since',
                                       min_time_str, 412)

    def test_handling_if_unmodified_since_header_200(self):
        """
        Test server handles If-Unmodified-Since header appropriately.
        If the file has not been modified since the time specified, 412 will be
        returned. Otherwise  200 with content.
        """
        print("If-Unmodified-Since - Testing 200 is returned if not modified since future.")
        prev_time = gmtime(time() + 600)
        prev_time_str = strftime('%a, %d %b %Y %H:%M:%S GMT', prev_time)
        verify_http_if_modified_header(self,
                                       'If-Unmodified-Since',
                                       prev_time_str, 200)

        print("If-Unmodified-Since - Testing if MAX date is accepted.")
        max_time = datetime(9999, 12, 31, 23, 59, 59, 999999)
        max_time_str = max_time.strftime('%a, %d %b %Y %H:%M:%S GMT')
        verify_http_if_modified_header(self,
                                       'If-Unmodified-Since',
                                       max_time_str, 200)


    def test_handling_if_modified_since_header_200(self):
        """
        Test server handles If-Modified-Since header appropriately.
        If the file has not been modified since the time specified, 304 will be
        returned. Otherwise  200 with content.
        """

        print("If-Modified-Since: Testing 200 is returned when modified time is in future.")
        prev_time = gmtime(time() - 10)
        prev_time_str = strftime('%a, %d %b %Y %H:%M:%S GMT', prev_time)
        verify_http_if_modified_header(self,
                                       'If-Modified-Since',
                                       prev_time_str, 200)

        print("If-Modified-Since: Testing if MIN date is accepted.")
        min_time_win = datetime(1601, 1, 1)
        min_time_str = min_time_win.strftime('%a, %d %b %Y %H:%M:%S GMT')
        verify_http_if_modified_header(self,
                                       'If-Modified-Since',
                                       min_time_str, 200)

        print("If-Modified-Since: Testing if Epoch start date is accepted.")
        min_time_unix = datetime(1970, 1, 1)
        min_time_str = min_time_unix.strftime('%a, %d %b %Y %H:%M:%S GMT')
        verify_http_if_modified_header(self,
                                       'If-Modified-Since',
                                       min_time_str, 200)

    def test_handling_ap_if_last_modified_200(self):
        """
        Test server handles X-IfLastModifiedTimestampGreater header.
        If the file has not been modified since the time specified, 304 will be
        returned. Otherwise 200 with content.
        """
        print("X-IfLastModifiedTimestampGreater: Testing 200 is returned when modified time is in past.")
        prev_time = calendar.timegm(gmtime()) - 600
        verify_http_if_modified_header(self,
                                       'X-IfLastModifiedTimestampGreater',
                                       posixtime_to_win_filetime(prev_time), 200)

        print("X-IfLastModifiedTimestampGreater: Testing if MIN date is accepted.")
        min_time_win = datetime(1601, 1, 1)
        min_timestamp = min_time_win.timestamp()
        verify_http_if_modified_header(self,
                                       'X-IfLastModifiedTimestampGreater',
                                       posixtime_to_win_filetime(min_timestamp), 200)

        print("X-IfLastModifiedTimestampGreater: Testing if Epoch start date is accepted.")
        min_time_unix = datetime(1970, 1, 1)
        min_timestamp = min_time_unix.timestamp()
        verify_http_if_modified_header(self,
                                       'X-IfLastModifiedTimestampGreater',
                                       posixtime_to_win_filetime(min_timestamp), 200)

    def test_handling_ap_if_last_modified_304(self):
        """
        Test server handles If-Modified-Since header appropriately.
        If the file has not been modified since the time specified, 304 will be
        returned.
        """
        print("X-IfLastModifiedTimestampGreater: Testing 304 is returned when modified time is in future.")
        later_time = calendar.timegm(gmtime()) + 600
        verify_http_if_modified_header(self,
                                       'X-IfLastModifiedTimestampGreater',
                                       posixtime_to_win_filetime(later_time), 304)

        print("X-IfLastModifiedTimestampGreater: Testing 304 is returned on FFFFFFFFFFFFFFFF date")
        max_time_stamp = 'FFFFFFFFFFFFFFFF'
        verify_http_if_modified_header(self,
                                        'X-IfLastModifiedTimestampGreater',
                                        max_time_stamp, 304)

        print("X-IfLastModifiedTimestampGreater: Testing 304 is returned on MAX date")
        max_time = datetime(9999, 12, 31, 23, 59, 59, 999999)
        max_time = max_time.replace(tzinfo=timezone.utc)
        max_time_stamp = max_time.timestamp()
        verify_http_if_modified_header(self,
                                        'X-IfLastModifiedTimestampGreater',
                                        posixtime_to_win_filetime(max_time_stamp), 304)

    def test_handling_if_modified_since_header_errors(self):
        """
        Test server handles If-Modified-Since http header errors appropriately.
        """
        with NamedTemporaryFile(dir=self.app_dir, prefix='autopilot.ini.') as tmpfile:
            later_time_str = strftime('%a, %d %b %Y %H:%M:%S WRN', gmtime(time()))

            file_name = PosixPath(tmpfile.name).name
            url = 'http://localhost:' + self.http_port + r'/d/app/' + file_name
            headers = {'If-Modified-Since': later_time_str}

            # if the content didn't change since time, we get 304
            resp = get(url, timeout=10, headers=headers)
            self.assertEqual(resp.status_code, 400,
                             "Expected HTTP Status Code 400")


"""
Helper Methods
"""


def verify_http_if_modified_header(self, header_str, date_str, expected_ret):
    """
    Test server handles If-Modified-Since http header errors appropriately.
    """
    with NamedTemporaryFile(dir=self.app_dir, prefix='autopilot.ini.') as tmpfile:
        file_name = PosixPath(tmpfile.name).name
        url = 'http://localhost:' + self.http_port + r'/d/app/' + file_name
        headers = {header_str: date_str}

        # if the content didn't change since time, we get 304
        resp = get(url, timeout=10, headers=headers)
        self.assertEqual(resp.status_code, expected_ret,
                         "Expected HTTP Status Code " + str(expected_ret))


def verify_dir_listing(self, resp, verify_file):
    self.assertEqual(resp.status_code, 200,
                     "Expected HTTP Status Code 200")

    lines = resp.text.splitlines()
    self.assertTrue(test_fsserver_headers(lines[0]),
                    "Did not return the correct dir listing header.")

    self.assertTrue(lines[1].startswith(r'.,4096,D'),
                    "Does not contain current dir info at second line.")

    self.assertTrue(lines[2].startswith(r'..,4096,D'),
                    "Does not contain parent dir info at third line.")

    file_found = False

    for line in lines:
        if line.startswith(verify_file):
            file_found = True
            break

    self.assertTrue(file_found,
                    verify_file + ' not found in dir listing.')
    print("RESPONSE :\n", resp.text)


def test_fsserver_headers(content):
    first_line = r'#Fields:filename,size,attributes,creationtime,lastaccesstime,lastwritetime'
    return content.startswith(first_line)


def posixtime_to_win_filetime(posix_time_secs):
    # Windows FILETIME = 100 nanoseconds intervals since Jan 1, 1601, 64-bit.
    # Posix time = seconds since Jan 1, 1970

    # Nano Seconds since Jan 1, 1601 to Jan 1, 1970
    epoch_diff_ns = 116444736000000000
    mtime_ns = posix_time_secs * 10000000
    win_file_time_ns = int(mtime_ns) + epoch_diff_ns

    hex_str = hex(win_file_time_ns)[2:].upper()

    # Prefix the hex string with 0 to fill 16 bytes
    file_time_str = ''.join((((16 - len(hex_str)) * '0'), hex_str))

    return file_time_str

def win_filetime_to_posixtime(win_filetime):
    # Windows FILETIME = 100 nanoseconds intervals since Jan 1, 1601, 64-bit.
    # Posix time = seconds since Jan 1, 1970
    ms = int(win_filetime, 16) / 10.

    return datetime.datetime(1601, 1, 1) + datetime.timedelta(microseconds=ms)

def get_fsserver_ini():
    """
    Return a fsserver.ini config object
    """
    config = ConfigParser()
    config.read_string("""
            [CherryPy]
            URLPathCaseSensitive=off

            [Certs]
            Directory=./certs
            SSLCert=CCPSCH010261732.crt
            SSLPrivateKey=CCPSCH010261732.key

            [SSLHosting]
            Port=6798

            [HTTPHosting]
            Port=6698

            [LINUX]
            AppDir=/tmp/ap/app
            DataDir=/tmp/ap/data
                      """)
    return config


if __name__ == '__main__':
    parser = ArgumentParser()
    parser.add_argument('-d')

    args = parser.parse_args()
    print(args)
    unittest.main()
