#!/usr/bin/env python3
# GetAndRunServices equivalent in Linux.
# Scheduled once every 10 mins
# 1. Get the Service List from DM - DONE
# 2. Get ServiceList ini from manifest server if there is a new version for
#    deployment
# 3. If PrepareList field has changed from last time, download the package from DS Server.
#     - Using HTTP(for now) or DataBus
# 4. If 'List' changed, then inform ServiceManager to start the new versions of services.
# 5. Drop the new servicelist.ini file for service manager
# 6. Update DM with the new service list version, even if no changes happened, but more the
#    6hrs passed since last update

import configparser
import sys
import os
import shutil
import tempfile
import filecmp
import inspect
import errno
import requests
import argparse
import statsd
from socket import gethostname


def _import_apcommon_path(apcommon):
    # Need to find and add relative ../apcommon dir to path to import apsdk et al
    # TODO: Find a better way to do this. Hint: Can also just wrap with start.sh
    common_path = os.path.realpath(
        os.path.abspath(
            os.path.join(
                os.path.split(
                    inspect.getfile(
                        inspect.currentframe()))[0], apcommon)))
    if common_path not in sys.path:
        sys.path.insert(0, common_path)

_import_apcommon_path('../apcommon')

from apsdk import dmclient
from apsdk import apck
from apsdk import env

# globals
dm_client = 0
metric = statsd.StatsClient(host='localhost', port=8125, prefix='gars')

# constants
#service_list_ini_path = '/usr/local/ap/data/ServiceManager/ServiceList.ini'
service_list_ini_path = os.path.join(env.get_data_root(),
                                     'servicemanager/ServiceList.ini')
copy_only_header = '#CopyOnly: '
data_copy_header = '#Data: '


def download_service_list_ini(dm_client):
    """
    Gets the service list from Device Manager
    """

    machine_name = gethostname()
    env_name = env.get_environment_name()

    cmd = 'GetServiceList -s ' + machine_name + ' -w ' + env_name

    t = metric.timer('GetServiceList.timing')
    t.start()
    resp = dm_client.send_request(cmd, '')
    t.stop()

    service_list = resp.text
    resp.connection.close()

    return service_list

@metric.timer('update_dm_service_status')
def update_dm_service_status(dm_client, list, prepare_list):
    # Command to update service list to DM
    cmd = 'ServiceStatus -serviceList ' + list + ' -prepareServiceList ' + prepare_list

    print ("Updating service status : ", cmd)

    t = metric.timer('ServiceStatus.timing')
    t.start()
    resp = dm_client.send_request(cmd, '')
    t.stop()

    return resp.text


def get_current_list(ini_config):
    return ini_config['ServiceManager']['List']


def get_prepare_list(ini_config):
    return ini_config['ServiceManager']['PrepareList']


def get_manifest_server_list(ini_config):
    if not ini_config.has_section('Manifest'):
        print ('ERROR: DM Reply does not have [Manifest] section :', ini_config)
    else:
        str = ini_config['Manifest']['ServerList']
        return str.split(':')


def get_env_name_from_svc_list_name(service_list_name, default_env=''):
    service_list_suffix = extract_service_list_suffix(service_list_name, False)

    # Dynamic service lists look like virtual ones, but they belong to
    # the current environment, not VE
    if service_list_suffix is not None and service_list_suffix.lower() == 'dynamic':
        return default_env

    return service_list_suffix


def get_pkgs_from_service_list(service_list_csv):
    pkg_list = list()
    try:
        with open(service_list_csv) as f:
            for line in f:
                # if line starts with #, we don't care about it yet
                if line.startswith('#'):
                    if line.startswith('#CopyOnly'):
                    # TODO:
                        pass
                    continue
                else:
                    # Line is of format:
                    #  APAMLocal.az_compute_2065832_AutoDeployment,0,0,,
                    items = line.split(sep=',')[0]
                    if len(items) > 4:
                        pkg_name = line.split(sep=',')[0]
                        pkg_list.append(pkg_name)

        return pkg_list

    except IOError as e:
        print ("ERROR: Service List csv could not be read :", service_list_csv)


def extract_service_list_suffix(service_list_name, bootstrap=False):
    if bootstrap:
        return 'bootstrap'

    items = service_list_name.split('.')
    count = len(items)
    if (count > 3) and items[count - 1] == 'csv':
        suffix = items[count - 2]
    else:
        suffix = None

    return suffix


def download_service_list_csv(manifest_server_list,
                              listen_port,
                              environment,
                              csv_name,
                              timeout):

    # Try from each server in the manifest server list
    for server in manifest_server_list:
        url = str.join('', ('http://', server, ':', str(listen_port),
                            '/ManifestServer?action=GetServiceList&environment=',
                            environment, '&name=', csv_name))

        # We will get a "301 - Moved Permanently" from Manifest server that
        # redirects us to the actual DynamicStorage url
        try:
            print ("Download Service List : ", url)
            resp = requests.get(url, allow_redirects=True, timeout=timeout)
            if resp.status_code == requests.codes.ok:
                content = resp.text
                resp.close()
                print ("Response : ", content)
                return content
            else:
                print ("Error downloading service list: ", resp.status_code, resp.text)

        except requests.exceptions.ConnectionError:
            print ("ConnectionError: ", url)
        except requests.exceptions.ConnectTimeout:
            print ("ConnectTimeout: ", url)
        except requests.exceptions.Timeout:
            print ("Timeout: ", url)


def get_apck_url(ip, listen_port, env_name, service_name):
    url = str.join('', ('http://', ip, ':', str(listen_port),
                   '//DynamicStorage/?action=DownloadStream&StreamName=stream://',
                   env_name, '/app/app.retail.amd64/',
                   service_name, '&version=apck'))
    return url


@metric.timer('download_service_pkg')
def download_service_pkg(service_name,
                         service_list_name,
                         image_root,
                         manifest_server_list,
                         listen_port,
                         timeout=4):

    env_name = get_env_name_from_svc_list_name(service_list_name)

    # Physical deployments do not have suffix .
    # TODO: Talk to AP folks and get the correct format. This doesn't sound right, even if it works
    if env_name is None:
        env_name = env.get_environment_name()

    # Create a directory with service_list_name to put the apck files
    service_file_name = os.path.basename(service_list_name)

    # Filename starts with a '.', omitting it
    if (service_file_name[0] is '.'):
        target_dir = os.path.join(image_root, service_file_name[1:])
    else:
        target_dir = os.path.join(image_root, service_file_name)

    if not os.path.exists(target_dir):
        shutil.os.mkdir(target_dir)

    # Try downloading from each server in the list
    download_size = 0
    for ip in manifest_server_list:
        apck_url = get_apck_url(ip, listen_port, env_name, service_name)
        print ("Downloading Service Pkg from : ", apck_url)

        # We will get a file to download from DynamicStorage server
        try:
            # Download as a stream as the size might become very large.
            # Stream download ensures memory doesn't blow out. Also lets us
            # separate handle header download and content download separately
            resp = None
            with metric.timer('pkg.download'):
                resp = requests.get(apck_url,
                                    stream=True,
                                    allow_redirects=True,
                                    timeout=timeout)
            metric.gauge('Anything', 1234.543)
            if resp.status_code != requests.codes.ok:
                print ("Error reading pkg file :", resp.status_code, resp.text)
                print ("Trying next server")
                resp.close()
                continue

            # print (
                #"Downloading",
                # resp.headers['X-FileSize'],
                #"bytes. CRC:",
                # resp.headers['X-FullCRC'])
            file_tmp = tempfile.NamedTemporaryFile(mode='wb',
                                                   dir=target_dir,
                                                   prefix=service_name,
                                                   suffix='.downloading',
                                                   delete=False)
            # Handle file download now.
            chunk_size = 1024 * 5
            for chunk in resp.iter_content(chunk_size):
                if chunk:
                    file_tmp.write(chunk)
                    file_tmp.flush()
                    print('.', end='')
            print()

            # release the connection back to the pool
            resp.close()

            download_size = file_tmp.tell()

            file_tmp.close()

            # If no content was received
            if download_size == 0:
                os.remove(file_tmp.name)
                return

            # Rename to original file name
            target_file_name = str.join('', (target_dir, '/', service_name, '.apck'))
            os.rename(file_tmp.name, target_file_name)

            return target_file_name, download_size

        except requests.exceptions.ConnectionError:
            print ("ConnectionError: ", apck_url)
            continue
        except requests.exceptions.ConnectTimeout:
            print ("ConnectTimeout: ", apck_url)
            continue
        except requests.exceptions.Timeout:
            print ("Timeout: ", apck_url)
            continue

    return None, 0

def parse_service_list_ini(service_list_ini):
    ini_config = configparser.ConfigParser()
    ini_config.read_string(service_list_ini)

    # Verify
    if not ini_config.has_section('ServiceManager'):
        print ('ERROR: DM Reply does not have [ServiceManager] section :',
               service_list_ini)
        return None

    if not ini_config.has_section('Manifest'):
        print ('ERROR: DM Reply does not have [Manifest] section :', service_list_ini)
        return None

    service_manager_dict = ini_config['ServiceManager']
    manifest_server_list = ini_config['Manifest']['ServerList'].split(':')

    if 'PrepareList' not in service_manager_dict:
        print ('ERROR: DM Reply does not have [ServiceManager.PrepareList] section :',
               service_list_ini)
        return None

    if 'List' not in service_manager_dict:
        print ('ERROR: DM Reply does not have [ServiceManager.List] section :',
               service_list_ini)
        return None

    return service_manager_dict, manifest_server_list


# Download the app packages inside the csv files given in the
def download_pkgs_in_csv(csv_file, manifest_server_list):
    print ("downloading pkgs from : ", csv_file)
    downloaded_pkgs = list()
    pkg_list = get_pkgs_from_service_list(csv_file)

    packed_images_dir = os.path.join(env.get_app_root(), 'packedimages')
    if not os.path.isdir(packed_images_dir):
        shutil.os.mkdir(packed_images_dir)

    num_pkgs_to_download = 1
    for pkg_name in pkg_list:
        print("Downloading : ", pkg_name)
        apck_file_name, size = download_service_pkg(pkg_name,
                                                    csv_file,
                                                    packed_images_dir,
                                                    manifest_server_list,
                                                    env.get_fsserver_http_port(),
                                                    timeout=20)
        if apck_file_name is None:
            metric.incr('pkgs.downloaded.error')
            print("Error Could not download :", pkg_name)
        else:
            metric.incr('pkgs.downloaded')
            print("Downloaded ", size, "bytes in :", apck_file_name)
            downloaded_pkgs.append(apck_file_name)

        # Hack during debugging. no need to download all the pkgs
        num_pkgs_to_download = num_pkgs_to_download - 1
        if num_pkgs_to_download == 0:
            break

    return downloaded_pkgs


def download_csv_from_list(csv_list, manifest_server_list):
    print ("downloading csvs from : ", csv_list)
    service_list_folder = os.path.join(env.get_app_root(), 'servicelists')
    if not os.path.isdir(service_list_folder):
        shutil.os.mkdir(service_list_folder)

    csv_list = csv_list.split(':')
    csv_file_list = list()

    for csv_name in csv_list:
        # Special case fo OM - Orchestration manager. We are going to ignore
        # it for now. We will tango with OM later.
        if csv_name.startswith('.OM.'):
            continue
        csv_content = download_service_list_csv(manifest_server_list,
                                                env.get_fsserver_http_port(),
                                                env.get_environment_name(),
                                                csv_name,
                                                timeout=15)
        if csv_content is None:
            print ("Error reading csv from manifest server : ", csv_name)
        else:
            csv_full_path = os.path.join(service_list_folder, csv_name)
            with open(csv_full_path, 'w') as f:
                csv_file_list.append(f.name)
                f.write(csv_content)

        # import ipdb; ipdb.set_trace()  # XXX BREAKPOINT
    return csv_file_list


def drop_service_list(service_list_ini):

    # Drop the ServiceList.ini for the Service Manager
    tmp_ini = tempfile.NamedTemporaryFile(mode='w', delete=False)
    tmp_ini.write(service_list_ini)

    # Drop if new and there is change from the current ServiceList.ini
    # TODO: Handle condition when service manager might be reading this
    #current_ini = '/usr/local/ap/data/servicemanager/new_servicelist.ini'
    current_ini = os.path.join(env.get_data_root(),
                               'servicemanager/new_servicelist.ini')
    if os.path.isfile(current_ini):
        if not filecmp.cmp(current_ini, tmp_ini.name):
            print ("Changes detected in new_servicelist.ini, writing : ", current_ini)
            os.remove(current_ini)
            shutil.move(tmp_ini.name, current_ini)
    else:
        print ("new_servicelist.ini not found, writing.")
        shutil.move(tmp_ini.name, current_ini)


# Extract a service package into the same directory
@metric.timer('pkg.extract')
def extract_service_package(apck_file_name):
    # TODO: MAJOR ERROR CHECKING NEEDED
    reader = apck.ApPackReader(apck_file_name)

    name_without_ext = os.path.splitext(os.path.basename(apck_file_name))[0]

    target_dir = os.path.join(os.path.dirname(apck_file_name),
                              name_without_ext)

    if not os.path.isdir(target_dir):
        shutil.os.mkdir(target_dir)

    reader.extract_all(target_dir)


def process_prepare_list(service_manager_dict, manifest_server_list):
    # If PrepareList changes, proceed downloading the packages. This is the
    # Pre-Fetching mechanism used by DM/GARS. We will later get a List entry
    # change that will show which service list should service manager be running.

    report_to_dm = False
    if service_manager_dict['PrepareList'] is not service_manager_dict['LastPrepareList']:
        report_to_dm = True
        prepare_list = service_manager_dict['PrepareList']
        downloaded_csv_files = download_csv_from_list(prepare_list, manifest_server_list)

        if len(downloaded_csv_files) > 0:
            # Get the changed application binaries
            pkg_lists_downloaded = []
            for csv_file in downloaded_csv_files:
                # Download all packages in the csv list
                downloaded_pkg_list = download_pkgs_in_csv(csv_file, manifest_server_list)
                if downloaded_pkg_list is not None:
                    pkg_lists_downloaded.append(downloaded_pkg_list)
                    print (len(downloaded_pkg_list),
                           "Packages downloaded.\n List: ",
                           downloaded_pkg_list)
                else:
                    print ('ERROR: pkgs could not be downloaded for:', csv_file)

            # Extract the downloaded packages.
            for pkg_list in pkg_lists_downloaded:
                for pkg in pkg_list:
                    # import ipdb; ipdb.set_trace()  # XXX BREAKPOINT
                    extract_service_package(pkg)
        else:
            print ('Error: No CSVs were downloaded')
    else:
        print ('PrepareList version did not change: ' + prepare_list)

    return report_to_dm


def parse_arguments():
    parser = argparse.ArgumentParser()
    parser.add_argument('--debug', '-d', action='store_true')
    return parser.parse_args()


def main():
    args = parse_arguments()

    # Get a new DM Client to talk to the lord(DM)
    dm_client = dmclient.DMClient('')
    report_to_dm = False

    # Get service list ini from DM
    service_list_ini = download_service_list_ini(dm_client)

    if (service_list_ini.startswith('No Response')):
        print ("No Reponse received from DM")
        exit(1)

    # If we have a new service list, then drop it for service manager
    drop_service_list(service_list_ini)

    # Parse service list
    #import ipdb; ipdb.set_trace()
    service_manager_dict, manifest_server_list = parse_service_list_ini(service_list_ini)
    if service_manager_dict is None:
        print ("Could not parse ServiceList ini returned by DM")
        exit(1)

    # TODO:
    # If List changes, then
    #    Wait for upto 15 mins for servicestatus.csv to reflect the changes.
    #      Service Manager is supposed to pick up the 'List' from the ServiceList.ini
    #      and run that version
    #    Send service status back to DM

    # Process the PrepareList in the Service Ini
    report_to_dm = process_prepare_list(service_manager_dict, manifest_server_list)

    # TODO: Update DM only if the service ver ini's 'LastReportTime' is >= 6hrs
    if args.debug is True:
        report_to_dm = True
    if report_to_dm:
        list = service_manager_dict['List']
        prepare_list = service_manager_dict['PrepareList']
        resp = update_dm_service_status(dm_client, list, prepare_list)
        if not resp.startswith('OK'):
            print ('ERROR: DM did not respond with OK to Service Status update :',
                   resp)
            exit(1)

if __name__ == '__main__':
    main()
