#!/usr/bin/env python3

import unittest
import cherrypy
import configparser
import filecmp
import getandrunservices as gars
import tempfile
import shutil
from random import randint
import sys
import os
import inspect


def _import_apcommon_path(apcommon):
    # Need to find and add relative ../apcommon dir to path to import apsdk et al
    # TODO: Find a better way to do this. Can just wrap with start.sh
    common_path = os.path.realpath(
        os.path.abspath(
            os.path.join(
                os.path.split(
                    inspect.getfile(
                        inspect.currentframe()))[0], apcommon)))
    if common_path not in sys.path:
        sys.path.insert(0, common_path)

_import_apcommon_path('../apcommon')
from apsdk import dmclient


class TestCherryServer(object):

    def __init__(self, dm_reply, manifest_redirect, ds_reply, apck_file_name):
        self.dm_reply = dm_reply
        self.manifest_redirect = manifest_redirect
        self.ds_reply = ds_reply
        self.apck_file_name = apck_file_name

    @cherrypy.expose
    def DMGateway(self):
        print('HEADERS: ', cherrypy.request.headers)
        cmd = cherrypy.request.headers['X-Apdm-Command']

        if str.startswith(cmd, 'GetServiceList -s'):
            return self.dm_reply
        elif str.startswith(cmd, 'ServiceStatus'):
            return 'OK'
        else:
            return 'ERROR'

    @cherrypy.expose
    def ManifestServer(self, **kwargs):
        raise cherrypy.HTTPRedirect(self.manifest_redirect, 301)

    @cherrypy.expose
    def DynamicStorage(self, **kwargs):
        if kwargs['version'] == 'apck':
            try:
                f = open(self.apck_file_name, 'rb')
                return f
            except EOFError as e:
                raise cherrypy.HTTPError(
                    status=404,
                    message='Empty File, could not be read:'
                    + self.apck_file_name + e)

            except IOError as e:
                raise cherrypy.HTTPError(
                    status=403,
                    message='File exists, but could not be read:'
                    + self.apck_file_name + e)
        else:

            return self.ds_reply


class TestIniParsing(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        print ("In Setup")
        cls.env = 'Autopilot'

        # Run the service on a random port
        cls.port = randint(13100, 13500)

        # Reply that should be returned from DMGateway
        cls.dm_reply = """
[ServiceManager]
List=AzureCloudvSwitch.az_compute_2060842_AutoDeployment_CL566092_432119.Autopilot-Autopilot-Test-VE.csv:.az_compute_2060842_AutoDeployment_CL564792_432119.Autopilot-AutopilotClient-Test-VE.csv
PrepareList=AzureCloudvSwitch.az_compute_2060842_AutoDeployment_CL566092_432119.Autopilot-Autopilot-Test-VE.csv:.az_compute_2060842_AutoDeployment_CL564792_432119.Autopilot-AutopilotClient-Test-VE.csv
LastList=AzureCloudvSwitch.az_compute_2053707_AutoDeployment_CL518554_432119.Autopilot-Autopilot-Test-VE.csv:.az_compute_2053707_AutoDeployment_CL520219_432119.Autopilot-AutopilotClient-Test-VE.csv
LastPrepareList=AzureCloudvSwitch.az_compute_2054480_AutoDeployment_CL523661_432119.Autopilot-Autopilot-Test-VE.csv:.az_compute_2053707_AutoDeployment_CL520219_432119.Autopilot-AutopilotClient-Test-VE.csv
LastReportTime=1900-01-01 00:00:00.000
StaticMachineFunction=AzureCloudvSwitch
StaticScaleunitId=100
DynamicMachineFunction=AzureCloudvSwitch
DynamicScaleunitId=100

[Manifest]
ServerList=1.2.3.4:127.0.0.1:10.208.84.45:10.203.244.113:10.203.244.19
PeerServerList=10.208.94.160:10.208.94.141:10.208.94.171:10.208.94.143:10.208.94.144:10.208.94.146:10.208.94.147:10.208.94.148:10.208.94.149:10.208.94.156:10.208.94.151:10.208.94.152:10.208.94.153:10.208.94.154:10.208.94.161:10.208.94.163:10.208.94.164:10.208.94.165:10.208.94.167:10.208.94.168:10.208.94.169:10.208.94.173:10.208.94.174:10.208.94.175:10.208.94.176:10.208.94.177:10.208.94.178:10.208.94.179:10.208.94.180:10.208.94.182:10.208.94.183:10.208.94.184:10.208.94.188

Manifest0=Distrib\Image\Autopilot\app_manifests\Manifestretail.amd64_AzureCloudvSwitch.csv
    """

        # Reply that should be returned from ManifestServer call
        cls.manifest_redirect = r'http://127.0.0.1:' + str(cls.port) + r'//DynamicStorage/?action=DownloadStream&StreamName=stream://' + \
            cls.env + '/app/servicelists/.Autopilot-AutopilotClient-Test-VE.csv' + '&version=az_compute_2065141_AutoDeployment_CL594637_432119'

        cls.ds_reply_service_list = """
#Version: 1.0
#Fields: services, interval, nohup, account, host
#CopyOnly:
#Data: ApSecurityGroups.az_compute_2065141_AutoDeployment:VE/Autopilot/AP/SecurityGroups
ActivateWindows.az_compute_2065832_AutoDeployment,3600,0,,
AlwaysOnMonitor.az_compute_2065832_AutoDeployment,120,0,,
APAMLocal.az_compute_2065832_AutoDeployment,0,0,,
    """

        # create a 2k dummy apck file
        #cls.apck_file = tempfile.NamedTemporaryFile(mode='wb', delete=False)
        #cls.apck_file.write(b'Hello\n')
        #cls.apck_file.truncate(1024 * 2)
        #cls.apck_file.close()
        cls.apck_file = "./test_data/ActivateWindows.az_compute_2065832_AutoDeployment.apck"

        config = {
            'server.socket_port': cls.port
        }

        cherrypy.config.update(config)
        cherrypy.tree.mount(TestCherryServer(cls.dm_reply,
                                             cls.manifest_redirect,
                                             cls.ds_reply_service_list,
                                             #cls.apck_file.name), '/', '')
                                             cls.apck_file), '/', '')
        cherrypy.engine.start()

    @classmethod
    def tearDownClass(cls):
        cherrypy.engine.exit()
        #os.remove(cls.apck_file)

    def test_dm_reply_is_parsed(self):
        dm_client = dmclient.DMClient(
            'http://127.0.0.1:' + str(self.port) + '/DMGateway', use_certs=False)

        # Make GARS request DM for service.ini
        ini_reply = gars.download_service_list_ini(dm_client)

        self.assertEqual(
            ini_reply,
            self.dm_reply,
            'dm reply text from gars was not equal to what dm replied.')

        ini_config = configparser.ConfigParser()
        ini_config.read_string(ini_reply)

        list = gars.get_current_list(ini_config)
        self.assertEqual(
            list,
            'AzureCloudvSwitch.az_compute_2060842_AutoDeployment_CL566092_432119.Autopilot-Autopilot-Test-VE.csv:.az_compute_2060842_AutoDeployment_CL564792_432119.Autopilot-AutopilotClient-Test-VE.csv')

        prepare_list = gars.get_prepare_list(ini_config)
        self.assertEqual(
            prepare_list,
            'AzureCloudvSwitch.az_compute_2060842_AutoDeployment_CL566092_432119.Autopilot-Autopilot-Test-VE.csv:.az_compute_2060842_AutoDeployment_CL564792_432119.Autopilot-AutopilotClient-Test-VE.csv')

        manifest_server_list = gars.get_manifest_server_list(ini_config)

        self.assertListEqual(manifest_server_list, ['1.2.3.4',
                                                    '127.0.0.1',
                                                    '10.208.84.45',
                                                    '10.203.244.113',
                                                    '10.203.244.19'],
                             "Manifest server IPs did not match")

        resp = gars.update_dm_service_status(dm_client, list, prepare_list)

        self.assertEqual(resp, 'OK', "DM Gateway did not return OK")

    def test_parse_service_list_name(self):
        suffix = gars.extract_service_list_suffix(
            ".ap_10_08_0_14453434_AutoDeployment_CL18437756.SUFFIX.csv")
        self.assertEqual(suffix, "SUFFIX")

        suffix = gars.extract_service_list_suffix(
            "DM.ap_10_08_0_14453434_AutoDeployment_CL18437756.SUFFIX.csv")
        self.assertEqual(suffix, "SUFFIX")

        suffix = gars.extract_service_list_suffix(
            "FCM.TESTap_10_08_0_14453434_AutoDeployment_CL18437756.Dynamic.csv")
        self.assertEqual(suffix, "Dynamic")

        suffix = gars.extract_service_list_suffix(
            "dm.ap_10_08_0_14453434_AutoDeployment_CL18437756.csv")
        self.assertEqual(suffix, None)

        suffix = gars.extract_service_list_suffix(
            "SUFFIX.csv")
        self.assertEqual(suffix, None)

    def test_service_list_download(self):
        csv_name = '.az_compute_2060842_AutoDeployment_CL564792_432119.Autopilot-AutopilotClient-Test-VE.csv'

        manifest_server_list = ['127.0.0.1',
                                '1.2.3.4',
                                '10.208.84.45',
                                '10.203.244.113',
                                '10.203.244.19']
        service_list_csv = gars.download_service_list_csv(
            manifest_server_list,
            self.port,
            self.env,
            csv_name,
            timeout=3)

        self.assertEqual(service_list_csv, self.ds_reply_service_list)

    # Verify that a list of apps is returned when a service_list csv file
    # containing services are passed
    def test_get_app_list_from_csv(self):

        service_list_csv = tempfile.NamedTemporaryFile(mode='w',
                                                       prefix='unittest',
                                                       delete=False)
        service_list_csv.write(self.ds_reply_service_list)
        service_list_csv.close()

        pkg_list = gars.get_pkgs_from_service_list(service_list_csv.name)

        self.assertListEqual(pkg_list, ['ActivateWindows.az_compute_2065832_AutoDeployment',
                                        'AlwaysOnMonitor.az_compute_2065832_AutoDeployment',
                                        'APAMLocal.az_compute_2065832_AutoDeployment'])

        os.remove(service_list_csv.name)
        """
        download new service ini
        if service_ini["LastReportTime"] > dm_report_threshold:
            report_to_dm = True

        if new_ini != old ini:
            report_to_dm = True
            for service_list_csv in prepare_list.split(':'):
                download service_list_csv into /usr/local/ap/app/servicelists
                    for app in servie_list_csv:
                        download the apck package
                        unzip it
            write new_ini to /usr/local/ap/data/servicemanager/new_servicelist.ini

        if report_to_dm:
            report to dm
        """

    # Test service package file download
    def test_download_and_extract_service_pkg(self):

        manifest_server_list = ['127.0.0.1',
                                '1.2.3.4',
                                '10.208.84.45',
                                '10.203.244.113',
                                '10.203.244.19']

        service_name = 'ActivateWindows.az_compute_2065832_AutoDeployment'
        service_list_name = '.az_compute_2060842_AutoDeployment_CL564792_432119.Autopilot-AutopilotClient-Test-VE.csv'
        target_dir = tempfile.mkdtemp(prefix='unit_test')
        apck_file_name, size = gars.download_service_pkg(service_name,
                                                         service_list_name,
                                                         target_dir,
                                                         manifest_server_list,
                                                         self.port,
                                                         timeout=4)

        self.assertEqual(apck_file_name,
                         target_dir + '/' + service_list_name[1:] + '/' + service_name + '.apck',
                         "gars didn't return the correct file name")

        self.assertTrue(os.path.isfile(apck_file_name),
                        "gars didn't write the file :" + apck_file_name)

        self.assertTrue(filecmp.cmp(apck_file_name, self.apck_file),
                        "apck file contents returned were different")

        gars.extract_service_package(apck_file_name)
        self.assertTrue(os.path.isfile(target_dir + '/' +
                                       service_list_name[1:] + '/' +
                                       service_name + '/start.bat'),
                        "Extracting start.bat failed.")

        # Cleanup
        os.remove(apck_file_name)
        shutil.rmtree(target_dir)


if __name__ == '__main__':
    # There is a known issue with urrlib that will throw warnings about
    # connection not being closed. So send warnings to be ignored
    unittest.main(warnings='ignore')
