import os
import inspect
import shutil
import sys
import apthelper

def _import_apcommon_path(apcommon):
    # Need to find and add relative ../apcommon dir to path to import apsdk et al
    # TODO: Find a better way to do this. Hint: Can also just wrap with start.sh
    common_path = os.path.realpath(
        os.path.abspath(
            os.path.join(
                os.path.split(
                    inspect.getfile(
                        inspect.currentframe()))[0], apcommon)))
    if common_path not in sys.path:
        sys.path.insert(0, common_path)

_import_apcommon_path('../apcommon')

from pyoverlayfs.exceptions import *
from pyoverlayfs.layer_assembly import LayerAssembly
from pysystemdwrapper.systemd_nspawn import NspawnContainer

class ServiceList:
    """ Creates an nspawn abstraction using overlayfs """

    def __init__(self, overlay, nspawn):
        self._overlay = overlay
        self._nspawn = nspawn

    @classmethod
    def create(cls, service_path):
        path = os.path.join(service_path, "root")
        overlay = cls.prepare_overlay(path)
        nspawn = NspawnContainer(path)
        return ServiceList(overlay, nspawn)
        
    @classmethod
    def load(cls, service_path):
        path = os.path.join(service_path, "root")
        overlay = load_overlay(path)
        nspawn = NspawnContainer(path)
        return ServiceList(overlay, nspawn)

    @property
    def path(self):
        return self._overlay.target_directory

    def rw_path(self):
        return self._overlay.directory

    def installed(self):
        return os.path.isfile(os.path.join(self.path, ".ap_install_complete"))

    def set_installed(self):
        open(os.path.join(self.path, ".ap_install_complete"), 'w').close()
        
    @classmethod
    def prepare_overlay(cls, path):
        rw_dir = os.path.join(os.path.join(path, "..", "._overlay"))

        if os.path.isfile(os.path.join(rw_dir, ".ap_install_complete")):
            print("Install already completed in this overlay")
        else:
            shutil.rmtree(rw_dir, ignore_errors=True)
            os.makedirs(rw_dir)
                
        print ("Mounting overlay folder: ", path)
        try:
            layer_assembly = LayerAssembly.create(directories=["/", rw_dir], target_directory=path)
            layer_assembly.mount()
        except MountFailureException as e:
            print ("ERROR: Unable to create overlay for %s: %s" % (path, e))
            raise e

        # resolv.conf of host system is copied over during nspawn.  Because this can be a symlink, we will
        # make sure we get the actual content of the file or DNS resolution will fail.
        resolv_conf_path = os.path.join(path, "etc/resolv.conf")
        try:
            os.remove(resolv_conf_path)
        except FileNotFoundError:
            pass
        shutil.copyfile("/etc/resolv.conf", resolv_conf_path, follow_symlinks=True)
        shutil.copy(apthelper.path(), path)

        return layer_assembly

    def run(self, cmd):
        # temporary delegation to nspawn.  TODO: model packages in service list
        self._nspawn.run(cmd.split())

    def teardown(self):
        print ("Unmounting overlay at %s" % self._overlay.target_directory)
        self._overlay.unmount()

    def purge_writes(self):
        print ("Purging writes made to: %s" % self.rw_path)
        shutil.rmtree(self.rw_path, ignore_errors=True)

    def apt_install(self, package, version=None):
        apt_command = "/apthelper.py --install " + package
        if version:
            apt_command = apt_command + " --version " + version

        try:
            self.run(apt_command)
        except Exception as e:
            print("Error from apt command %s: %s " % (apt_command, e))
            return False

        return True

    def apt_remove(self, package):
        apt_command = "/apthelper.py --remove " + package
        try:
            self.run(apt_command)
        except Exception as e:
            print("Error from apt command %s: %s " % (apt_command, e))
            return False

        return True
