#!/usr/bin/env python3

import os
import shutil
import configparser
import subprocess
import inspect
import sys
import apthelper
from jinja2 import Template
from servicelist import ServiceList

def _import_apcommon_path(apcommon):
    # Need to find and add relative ../apcommon dir to path to import apsdk et al
    # TODO: Find a better way to do this. Hint: Can also just wrap with start.sh
    common_path = os.path.realpath(
        os.path.abspath(
            os.path.join(
                os.path.split(
                    inspect.getfile(
                        inspect.currentframe()))[0], apcommon)))
    if common_path not in sys.path:
        sys.path.insert(0, common_path)

_import_apcommon_path('../apcommon')

from apsdk import env
from apsdk import svclists
from pysystemdwrapper.unitfiles import UnitFile
from pysystemdwrapper.units import Unit

def touch_servicestatus():
    csv = '/usr/local/ap/data/servicemanager/servicestatus.csv'
    localwatchdog = '/usr/local/ap/data/watchdogoutput/localwatchdogstatus.csv'

    if os.path.exists(csv):
        print ('Updating time on : ', csv)
        os.utime(csv)
        os.utime(localwatchdog)
    else:
        print ("ERROR: File not found :", csv)


# Activate the service list. Will contain a list of CSVs to be activated.
# Delimited by :
def activate_service_list(activate_list):
    csv_list = activate_list.split(':')

    for csv_name in csv_list:
        # Ignore orchestration manager for now
        if csv_name.startswith('.OM'):
            continue

        # Ignore the APTools too since they are for Windows
        if csv_name.startswith('.ap_'):
            continue

        print ("Activate Service List :", activate_list)
        if activate_service_list_csv(csv_name) is False:
            print("Error activating Service List CSV :", csv_name)
            return False

    return True


def activate_service_list_csv(service_list_name):
    pkg_list = svclists.get_pkg_list(service_list_name)
    if pkg_list is None:
        print ("No packages in service list csv :", service_list_name)
        return False

    service_list_pkg_path = svclists.get_pkg_root(service_list_name)

    for pkg_name in pkg_list:
        pkg_dir = os.path.join(service_list_pkg_path, pkg_name)

        if not os.path.isdir(pkg_dir):
            print ("ERROR: package path does not exist :", pkg_dir)
            #continue
            return False

        if not validate_package(pkg_dir):
            print ("ERROR: Package validation failed. Ignoring pkg:", pkg_name)
            pkg_list.remove(pkg_name)
            continue

        try:
            service_list = ServiceList.create(service_list_pkg_path)
        except Exception as e:
            print("Exception in service list creation: ", e)
            return False

        if not service_list.installed():
            print("Service list not installed.  Processing requirements now")
            if not process_requirements(pkg_dir, service_list):
                print ("ERROR: Requirements could not be satisfied. Ignoring pkg :", pkg_name)
                pkg_list.remove(pkg_name)
                service_list.teardown()
                continue
        else:
            print("Service list was already installed.")

    print ("Requirement processing complete, activating packages now.")

    for pkg_name_full in pkg_list:
        # Pkg name is of format <pkg_name>.<version>
        # e.g. AlwaysOnMonitor.az_compute_2065832_AutoDeployment
        pkg_name = pkg_name_full.split('.')[0]
        pkg_dir = os.path.join(service_list.path, pkg_name_full)

        if not service_list.installed():
            print ("Copying %s into %s in overlay" % (pkg_dir, service_list.path))

            shutil.copytree(os.path.join(service_list_pkg_path, pkg_name_full),
                            os.path.join(service_list.path, pkg_name))
            service_list.set_installed()
        else:
            print("Service list was already installed.  Skipping %s copy" % pkg_name_full)

        if not activate_package(pkg_name, service_list):
            print ("ERROR: Package could not be activated :", pkg_name_full)
            return False
            # TODO: Set error to propagate to state


def validate_package(pkg_dir):
    print ("Validating :", pkg_dir)
    for f in ['start.sh']:
        file = os.path.join(pkg_dir, f)
        if not os.path.isfile(file):
            print ("Package missing required file:", file)
            return False
    return True


def activate_package(pkg_name, service_list):
    print ("Activating: ", pkg_name)

    pkg_dir = os.path.join(service_list.path, pkg_name)

    # Run the install script, if available
    if not execute_install_sh(pkg_name, service_list):
        return False

    # Create pkg_name symlink in /usr/local/ap/app/<pkg_name> to the
    pkg_app_link = os.path.join(env.get_app_root(), pkg_name)
    try:
        print ("Creating symlink :", pkg_app_link)
        os.symlink(pkg_dir, pkg_app_link)
    except Exception as e:
        print ("Error creating symlink :", str(e))
        print ("Failed Activation: This package was not deactivated properly")
        return False

    # Render systemd-nspawn systemd unit
    start_sh_path =  os.path.join(pkg_dir, 'start.sh')
    app_name = pkg_name.split('.')[0]
    service_name = app_name + '.service'
    service_template_path = os.path.join(os.path.dirname(os.path.realpath(__file__)),
                                         'app.service.template')
    unit_file_path = os.path.join(service_list.path, service_name)
    print ("Loading template ", service_template_path)
    with open(service_template_path, 'r') as template_file:
        template_content = template_file.read()

    service_template = Template(template_content)
    unit_file_contents = service_template.render(app_name=app_name, service_path=service_list.path)
    with open(unit_file_path, "w") as unit_file:
        unit_file.write(unit_file_contents)

    print ("Linking unit file: ", unit_file_path)
    UnitFile.link(unit_file_path)
    app_unit = Unit.find(service_name)

    print ("Starting unit : ", app_unit.name)
    app_unit.start()

    print ("Succesfully Activated:", app_name)
    return True


def execute_install_sh(pkg_name, service_list):
    # Run the install script, if available
    install_sh = os.path.join(pkg_name, 'install.sh')
    if os.path.isfile(os.path.join(service_list.path, install_sh)):
        try:
            print ("Found install.sh in the package, running:", install_sh)
            # Start a new shell to execute script and run in it's pkg dir
            service_list.run(install_sh)
        except subprocess.CalledProcessError as e:
            print ("Error running install.sh:", install_sh, str(e))
            return False
        print ("Successfully ran install.sh")
    else:
        print ("No install.sh found.  Continuing", os.path.join(service_list.path, install_sh))
    return True

def process_requirements(pkg_full_path, service_list):
    req_ini_path = os.path.join(pkg_full_path, 'requirements.ini')
    if not os.path.isfile(req_ini_path):
        return True

    print ("Processing requirements for", req_ini_path)
    reqs = configparser.ConfigParser(allow_no_value=True)
    try:
        reqs.read(req_ini_path)
    except Exception as e:
        print ("Error Processing ini:", req_ini_path, str(e))
        return False

    # install packages
    for apt_pkg_name in reqs['Apt-Get-Packages']:
        apt_pkg_ver = reqs['Apt-Get-Packages'][apt_pkg_name]

        if not service_list.apt_install(apt_pkg_name, apt_pkg_ver):
            print ("Installing package failed from requirements.ini :", req_ini_path)
            return False

    return True

def apt_get_install(pkg_name, pkg_ver=None):
    apt_cache = apt.cache.Cache()

    # TODO: Handle dependencies
    if pkg_name not in apt_cache:
        # Let's update the cache and try once more
        # Updating cache seems to be unreliable. So will not do so.
        #try:
            #apt_cache.update()
        #except apt.cache.LockFailedException as e:
            #print("Could not refresh apt cache: ", str(e))
        #else:
            #if pkg_name not in apt_cache:
        print ("No such package exists :", pkg_name)
        return False

    apt_pkg = apt_cache[pkg_name]

    # Pin the specific version available
    if pkg_ver is not None:
        if pkg_ver in apt_pkg.versions:
            apt_pkg.candidate = apt_pkg.versions[pkg_ver]
        else:
            print ("Requested version could not be found for pkg", pkg_name, pkg_ver)
            return False

    # TODO: Handle dependencies
    if apt_pkg.is_installed:
        print ('{0} already installed'.format(pkg_name))
        return True
    else:
        apt_pkg.mark_install(auto_fix=True)

        try:
            print ("Installing ", pkg_name, pkg_ver)
            apt_cache.commit()
        except Exception as arg:
            print ("ERROR: Package installation failed :", str(arg))
            return False
        else:
            return True


def apt_get_remove(pkg_name):
    apt_cache = apt.cache.Cache()

    # TODO: Handle dependencies
    if pkg_name not in apt_cache:
        # Let's update the cache and try once more
        apt_cache.update()
        if pkg_name not in apt_cache:
            print ("No such package exists :", pkg_name)
            return False

    apt_pkg = apt_cache[pkg_name]

    if not apt_pkg.is_installed:
        print ('{0} not installed'.format(pkg_name))
        return 0
    else:
        apt_pkg.mark_delete(purge=True)

        try:
            apt_cache.commit()
        except Exception as arg:
            print ("ERROR: Package installation failed :", str(arg))
            return False
        else:
            return True


def check_service_switch(new_ini_data, current_ini_data):
    #print ("Checking: ", new_ini_data)
    #if not os.path.isfile(new_ini_data):
        #print("File not found, exitting :", new_ini_data)
        #return None, None

    ini = configparser.ConfigParser()
    try:
        ini.read_string(new_ini_data)
    except Exception as e:
        print ("Error Processing ini:", new_ini_data, str(e))
        return None, None

    if not ini.has_section('ServiceManager'):
        print ("Error processing new_ini_data, ServiceManager section not found.")
        return None, None

    new_list = ini['ServiceManager']['List']

    if current_ini_data is None:
        print("No current servicemap ini file found. This is the first version.")
        return new_list, None

    # No need to check again since we saved this after validating last time
    ini.read_string(current_ini_data)
    current_list = ini['ServiceManager']['List']

    if new_list != current_list:
        return new_list, current_list

    return None, None


def deactivate_service_list(deactivate_list):
    csv_list = deactivate_list.split(':')

    for csv_name in csv_list:
        # Ignore orchestration manager for now
        if csv_name.startswith('.OM'):
            continue

        # Ignore the APTools too since they are for Windows
        if csv_name.startswith('.ap_'):
            continue

        pkg_list = svclists.get_pkg_list(csv_name)
        if pkg_list is None:
            print("No packages found in ", csv_name)
            continue

        for pkg_name_with_ver in pkg_list:
            if not deactivate_package(pkg_name_with_ver):
                print("Error: Deactivation failed for CSV")
                return False
    return True

def deactivate_package(pkg_name_with_ver):
    app_name = pkg_name_with_ver.split('.')[0]
    print("Deactivating :", app_name)

    # Stop the systemd service
    # TODO: Should send signal, wait, check status periodically and then kill if needed.
    try:
        subprocess.check_call(["systemctl", "stop", app_name])
    except subprocess.CalledProcessError as e:
        print ("Error stopping service:", app_name, str(e))
        #return False

    # Remove the systemd unit file
    service_file = os.path.join('/lib/systemd/system/', app_name + ".service")
    if os.path.isfile(service_file):
        print ("Removing Service File :", service_file)
        os.remove(service_file)
    else:
        print ("Systemd service file was not found for the service")

    # Remove the symlink
    app_link = os.path.join(env.get_app_root(), app_name)
    if os.path.islink(app_link):
        print ("Removing Link File :", app_link)
        os.unlink(app_link)
    else:
        print ("Application Symbolic Link was not found for the service:", app_link)

    print ("Deactivated: ", app_name)
    return True

def load_service_list_ini(service_list_ini):
    print ("Loading: ", service_list_ini)
    if not os.path.isfile(service_list_ini):
        print("File not found, exitting :", service_list_ini)
        return None

    with open(service_list_ini, "r") as f:
        return f.read()

def main():
    touch_servicestatus()

    # Check if the service list should be switched
    new_servicelist_ini = os.path.join(env.get_data_root(), 'servicemanager/new_servicelist.ini')
    current_servicelist_ini = os.path.join(env.get_data_root(), 'servicemanager/current_servicelist.ini')

    # Load into a string so that we don't get into races with gars during processing
    new_ini_data = load_service_list_ini(new_servicelist_ini)
    current_ini_data = load_service_list_ini(current_servicelist_ini)

    activate_list, deactivate_list = check_service_switch(new_ini_data,
                                                          current_ini_data)
    # If changed, deactivate the old version
    if deactivate_list is not None:
        print("Deactivating packages from list: ", deactivate_list)
        if not deactivate_service_list(deactivate_list):
            print("Deactivation failed for service list. Exitting without switching service")
            return

    if activate_list is None:
        print ("No new list to activate.")
        return

    # Activate the new version of the service
    if not activate_service_list(activate_list):
        print("Activation failed for service list. Exitting without switching service")
        # TODO: We should notify gars via an error that switch has not happened.
        return

    # We have succesfully activated, now write this as the current service list
    if os.path.isfile(current_servicelist_ini):
    	os.remove(current_servicelist_ini)

    with open(current_servicelist_ini, "w") as f:
        f.write(new_ini_data)
    print("Processed and new current service list is written:", current_servicelist_ini)

if __name__ == '__main__':
    main()
