#!/usr/bin/env python3

import unittest
import cherrypy
import configparser
import filecmp
import tempfile
import shutil
import os
import apt
import servicemanager
from servicelist import ServiceList

class TestPackageManagement(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.service_list_root =  "/tmp/ap_integration_tests"
        try:
            os.makedirs(cls.service_list_root)
        except:
            pass
        cls.servicelist = ServiceList.create(cls.service_list_root)

    @classmethod
    def tearDownClass(cls):
        cls.servicelist.purge_writes()
        cls.servicelist.teardown()
        pass

    def test_apt_pkg_install(self):
        pkg = 'python3-tz'
        res = self.servicelist.apt_install(pkg)
        self.assertTrue(res, '{0} could not be installed'.format(pkg))

        apt_cache = apt.cache.Cache(rootdir=self.servicelist.path)
        apt_pkg = apt_cache[pkg]
        self.assertTrue(apt_pkg.is_installed, 'Package was not installed')

    def test_apt_pkg_install_existing(self):
        pkg = 'python3-tz'
        res = self.servicelist.apt_install(pkg)
        self.assertTrue(res, '{0} could not be installed'.format(pkg))

        apt_cache = apt.cache.Cache(rootdir=self.servicelist.path)
        apt_pkg = apt_cache[pkg]
        self.assertTrue(apt_pkg.is_installed, 'Package was not installed')

    def test_apt_pkg_install_incorrect(self):
        pkg = 'python3-NON_EXISTINGPKG'
        res = self.servicelist.apt_install(pkg)
        self.assertFalse(res, '{0} should not be accepted for installation'.format(pkg))

    def test_apt_pkg_remove(self):
        pkg = 'python3-tz'
        res = self.servicelist.apt_remove(pkg)
        self.assertTrue(res, '{0} could not be un-installed'.format(pkg))

        apt_cache = apt.cache.Cache(rootdir=self.servicelist.path)
        apt_pkg = apt_cache[pkg]

        self.assertTrue(apt_pkg.is_installed is False, 'Package was not removed')

    def test_apt_pkg_remove_non_existing(self):
        pkg = 'python3-tz'
        res = self.servicelist.apt_remove(pkg)
        self.assertEquals(res, 0, '{0} could not be un-installed'.format(pkg))

        apt_cache = apt.cache.Cache(rootdir=self.servicelist.path)
        apt_pkg = apt_cache[pkg]
        self.assertTrue(apt_pkg.is_installed is False, 'Package was not removed')


class TestProcessPackageRequirements(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.test_data_path = os.path.join(
            os.path.dirname(os.path.realpath(__file__)),
            'unittest'
        )

        if not os.path.isdir(cls.test_data_path):
            shutil.os.mkdir(cls.test_data_path)

        cls.valid_pkg_dir = tempfile.mkdtemp(dir=cls.test_data_path,
                                             prefix='unit_test')
        cls.pkg_dir_no_start_sh = tempfile.mkdtemp(dir=cls.test_data_path,
                                                   prefix='unit_test')

        cls.service_list_root =  "/tmp/ap_integration_tests"
        try:
            os.makedirs(cls.service_list_root)
        except:
            pass
        cls.servicelist = ServiceList.create(cls.service_list_root)

        install_sh = os.path.join(cls.valid_pkg_dir, 'install.sh')
        with os.fdopen(os.open(install_sh, os.O_WRONLY | os.O_CREAT), 'w') as f:
            f.write("""
#!/usr/bin/env bash
touch /install_executed
                    """)

        req_ini = os.path.join(cls.valid_pkg_dir, 'requirements.ini')
        with open(req_ini, 'w+') as f:
            f.write("""
[Apt-Get-Packages]
python3-cherrypy3
nginx=1.6.2-5ubuntu3.1
python3-flask
                    """)

        systemd_unit = os.path.join(cls.valid_pkg_dir, 'systemd.unit')
        start_sh = os.path.join(cls.valid_pkg_dir, 'start.sh')
        with open(systemd_unit, 'w+') as f:
            f.write("""
[Unit]
Description=AP Test Service

[Service]
Type=simple
"""
                    +
                    "ExecStart=" + start_sh
                    +
                    """
Restart=always
RestartSec=1min
""")

        #import ipdb; ipdb.set_trace()  # XXX BREAKPOINT

        with open(start_sh, 'w+') as f:
            f.write("""
#!/usr/bin/env bash
echo "BLAH BLAH BLAHstart_sh, 0755"
                    """)
        os.chmod(start_sh, 755 )

        # Hack the work that prepares the service list container (TODO: refactor)
        shutil.copytree(cls.valid_pkg_dir, os.path.join(cls.servicelist.path, os.path.basename(cls.valid_pkg_dir)))
        
    @classmethod
    def tearDownClass(cls):
        #shutil.rmtree(cls.valid_pkg_dir)
        cls.servicelist.purge_writes()
        cls.servicelist.teardown()
        shutil.rmtree(cls.pkg_dir_no_start_sh)

    def test_activate_package(self):
        res = servicemanager.activate_package(os.path.basename(self.valid_pkg_dir),
                                              self.servicelist)
        self.assertTrue(res, "Activation failed for pkg")
        self.assertTrue(os.path.isfile(
            os.path.join(self.servicelist.path, 'install_executed')
        ), "install.sh did not execute.")

    def test_validate_package(self):
        self.assertTrue(servicemanager.validate_package(self.valid_pkg_dir),
                        'Validation did not pass for a good package')

        self.assertFalse(servicemanager.validate_package(self.pkg_dir_no_start_sh),
                        'Validation did not fail for a bad package')

    def test_process_requirements(self):
        res = servicemanager.process_requirements(self.test_data_path, self.servicelist)


if __name__ == '__main__':
    # There is a known issue with urrlib that will throw warnings about
    # connection not being closed. So send warnings to be ignored
    unittest.main(warnings='ignore')
